#include <iostream>
#include <chrono>

#include <opencv2/core.hpp>
#include <opencv2/videoio.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>

void test_iterate(cv::Mat& m);

int main(int argc, char* argv[])
{
	cv::VideoCapture capture;

	bool success = true;
	if (argc < 2)
	{
		success = capture.open(0);
	}
	else
	{
		success = capture.open(argv[1]);
	}
	if (!success)
	{
		std::cerr << "Unable to open video capture" << std::endl;
		return 0;
	}

	cv::namedWindow("Input");
	cv::namedWindow("Processed");

	cv::Mat frame;
	while (capture.read(frame))
	{
		test_iterate(frame);

		cv::imshow("Input", frame);

		char key = static_cast<char>(cv::waitKey(1));
		if (key == 27) { std::cout << "Exiting..." << std::endl; break; }
	}

	return 0;
}

// version 1: iterators
// http://docs.opencv.org/2.4/doc/tutorials/core/how_to_scan_images/how_to_scan_images.html#howtoscanimagesopencv
void iterate_v1(cv::Mat& m, unsigned char threshold)
{
	for (cv::MatIterator_<cv::Vec3b> it = m.begin<cv::Vec3b>(), end = m.end<cv::Vec3b>(); it != end; ++it)
	{
		(*it)[0] = (*it)[0] > threshold ? 255 : 0;
		(*it)[1] = (*it)[1] > threshold ? 255 : 0;
		(*it)[2] = (*it)[2] > threshold ? 255 : 0;
	}
}


// version 2: .at
void iterate_v2(cv::Mat& m, unsigned char threshold)
{
	for (int i = 0; i < m.rows; ++i)
	{
		for (int j = 0; j < m.cols; ++j)
		{
			cv::Vec3b pixel = m.at<cv::Vec3b>(i, j);
			pixel[0] = pixel[0] > threshold ? 255 : 0;
			pixel[1] = pixel[1] > threshold ? 255 : 0;
			pixel[2] = pixel[2] > threshold ? 255 : 0;
		}
	}
}

// version 3.1: pointers
// http://docs.opencv.org/2.4/doc/tutorials/core/how_to_scan_images/how_to_scan_images.html#howtoscanimagesopencv
void iterate_v3_1(cv::Mat& m, unsigned char threshold)
{
	int nChannels = m.channels();
	int nRows = m.rows, nCols = m.cols * nChannels;
	if (m.isContinuous())
	{
		nCols *= nRows;
		nRows = 1;
	}
	for (int i = 0; i < nRows; ++i)
	{
		uchar* p = m.ptr<uchar>(i);
		for (int j = 0; j < nCols; ++j)
		{
			p[j] = p[j] > threshold ? 255 : 0;
		}
	}
}

// version 3.2
void iterate_v3_2(cv::Mat& m, unsigned char threshold)
{
	CV_Assert(m.isContinuous());
	int nChannels = m.channels();
	int nRows = m.rows, nCols = m.cols * nChannels;
	uchar* p = m.data;
	for (int i = 0; i < nCols*nRows; ++i)
	{
		*p++ = *p > threshold ? 255 : 0;
	}
}

// version 4: foreach
void iterate_v4(cv::Mat& m, unsigned char threshold)
{
	m.forEach<cv::Vec3b>([&threshold](cv::Vec3b& p, const int * position) -> void {
		p[0] = p[0] > threshold ? 255 : 0;
		p[1] = p[1] > threshold ? 255 : 0;
		p[2] = p[2] > threshold ? 255 : 0;
	});
}

// version 5: OpenCV
void iterate_v5(cv::Mat& m, unsigned char threshold)
{
	cv::threshold(m, m, threshold, 255.0, cv::THRESH_BINARY);
}

void test_iterate(cv::Mat& m)
{
	CV_Assert(m.type() == CV_8UC3);

	unsigned char threshold = 128;
	std::vector<std::function<void(cv::Mat&, unsigned char)>> iterate_versions = {
		iterate_v1,
		iterate_v2,
		iterate_v3_1,
		iterate_v3_2,
		iterate_v4,
		iterate_v5
	};
	std::vector<std::string> functionString = {
		"Version 1, iterator:\t",
		"Version 2, .at:\t\t",
		"Version 3.1, pointer:\t",
		"Version 3.2, pointer:\t",
		"Version 4, foreach:\t",
		"Version 5, OpenCV:\t",
	};

	cv::Mat m_clone;
	for (size_t i = 0; i < iterate_versions.size(); ++i)
	{
		auto t_start = std::chrono::high_resolution_clock::now();
		m_clone = m.clone();

		iterate_versions[i](m_clone, threshold);

		auto t_end = std::chrono::high_resolution_clock::now();
		std::cout << functionString[i]
			<< std::chrono::duration_cast<std::chrono::microseconds>(t_end - t_start).count() << " usec, "
			<< std::chrono::duration_cast<std::chrono::milliseconds>(t_end - t_start).count() << " msec"
			<< std::endl;
	}
	std::cout << std::endl;

	cv::imshow("Processed", m_clone);
}
